# Copyright 2019 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stark.utils import create_output_string_lemma


class Tree(object):
    def __init__(self, index, form, lemma, upos, xpos, deprel, head, feats_detailed, misc, document, summary):

        feats_dict = summary.feats_dict

        if not hasattr(self, 'feats'):
            self.feats_detailed = {}

        if form not in document.form_dict:
            document.form_dict[form] = form
        self.form = document.form_dict[form]
        if lemma not in document.lemma_dict:
            document.lemma_dict[lemma] = lemma
        self.lemma = document.lemma_dict[lemma]
        if upos not in document.upos_dict:
            document.upos_dict[upos] = upos
        self.upos = document.upos_dict[upos]
        if xpos not in document.xpos_dict:
            document.xpos_dict[xpos] = xpos
        self.xpos = document.xpos_dict[xpos]
        if deprel not in document.deprel_dict:
            document.deprel_dict[deprel] = deprel
        self.deprel = document.deprel_dict[deprel]
        for feat in feats_detailed.keys():
            if feat not in feats_dict:
                feats_dict[feat] = {}
            if next(iter(feats_detailed[feat])) not in feats_dict[feat]:
                feats_dict[feat][next(iter(feats_detailed[feat]))] = next(iter(feats_detailed[feat]))
            if feat not in self.feats_detailed:
                self.feats_detailed[feat] = {}
            self.feats_detailed[feat][next(iter(feats_detailed[feat]))] = (
                feats_dict)[feat][next(iter(feats_detailed[feat]))]

        self.feats = {k: list(v.keys())[0] for k, v in self.feats_detailed.items()}
        if misc not in document.misc_dict:
            document.misc_dict[misc] = misc
        self.misc = document.misc_dict[misc]

        self.parent = head
        self.children = []
        self.children_split = -1
        self.conll = None

        self.index = index

        # for caching answers to questions
        self.cache = {}

    def add_child(self, child):
        self.children.append(child)

    def set_parent(self, parent):
        self.parent = parent

    def get_unigrams(self, create_output_strings):
        unigrams = [Tree._generate_key(self, create_output_strings, print_lemma=False)[1]]
        for child in self.children:
            unigrams += child.get_unigrams(create_output_strings)
        return unigrams

    def add_conll_sentence(self, conll):
        self.conll = conll

    def get_root(self):
        """
        Get root of a node.
        :return:
        """
        root = self
        while root.parent is not None:
            root = root.parent

        return root

    @staticmethod
    def _generate_key(node, create_output_strings, print_lemma=True):
        array = [[create_output_string(node) for create_output_string in create_output_strings]]
        if create_output_string_lemma in create_output_strings and print_lemma:
            key_array = [[create_output_string(
                node) if create_output_string != create_output_string_lemma else create_output_string(node) for
                          create_output_string in create_output_strings]]
        else:
            key_array = array
        if len(array[0]) > 1:
            key = '&'.join(key_array[0])
        else:
            key = key_array[0][0]

        return array, key




# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Summary(object):
    """
    A class that is used to store results of processing.
    """
    def __init__(self):
        self.feats_dict = {}
        self.query_trees = None
        self.samples = []
        self.corpus_size = 0
        self.unigrams = {}
        self.representation_trees = {}
        self.max_tree_size = 0

    def set_query_trees(self, query_trees):
        """
        A function that sets query trees when they are needed.
        :param query_trees:
        :return:
        """
        self.query_trees = query_trees

    def get_summary_data(self):
        """
        A function that returns summary data used for storing cache.
        :return:
        """
        return (self.representation_trees, self.unigrams, self.corpus_size, self.feats_dict,
                self.samples, self.max_tree_size, self.query_trees)

    @classmethod
    def create_summary_from_cache(cls, sum_data):
        """
        A function that forms summary from cache.
        :param sum_data:
        :return:
        """
        s = cls()
        s.representation_trees, s.unigrams, s.corpus_size, s.feats_dict, s.samples, s.max_tree_size, s.query_trees = (
            sum_data)
        return s

    # def get_size_representation_trees(self):
    #     size = 0
    #     for represetation_tree_k, represetation_tree_v in self.representation_trees.items():
    #         size += sys.getsizeof(represetation_tree_k)
    #         size += sys.getsizeof(represetation_tree_v['number'])
    #         size += sys.getsizeof(represetation_tree_v['object'])
    #         size += represetation_tree_v['object'].get_size()
    #
    #     return size



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from stark.resources.constants import UNIVERSAL_FEATURES


def add_node(tree):
    """
    Adds node to a tree.
    :param tree:
    :return:
    """
    if 'children' in tree:
        tree['children'].append({})
    else:
        tree['children'] = [{}]


def tree_grow(orig_tree):
    """
    Walks over all nodes in tree and add a node to each possible node.
    :param orig_tree:
    :return:
    """
    new_trees = []
    new_tree = copy.deepcopy(orig_tree)
    add_node(new_tree)
    new_trees.append(new_tree)
    if 'children' in orig_tree:
        children = []
        for child_tree in orig_tree['children']:
            children.append(tree_grow(child_tree))
        for i, child in enumerate(children):
            for child_res in child:
                new_tree = copy.deepcopy(orig_tree)
                new_tree['children'][i] = child_res
                new_trees.append(new_tree)

    return new_trees


def compare_trees(tree1, tree2):
    """
    Compares two trees and returns True, when they match.
    :param tree1:
    :param tree2:
    :return:
    """
    if tree1 == {} and tree2 == {}:
        return True

    if 'children' not in tree1 or 'children' not in tree2 or len(tree1['children']) != len(tree2['children']):
        return False

    children2_connections = []

    for child1_i, child1 in enumerate(tree1['children']):
        child_duplicated = False
        for child2_i, child2 in enumerate(tree2['children']):
            if child2_i in children2_connections:
                pass
            if compare_trees(child1, child2):
                children2_connections.append(child2_i)
                child_duplicated = True
                break
        if not child_duplicated:
            return False

    return True


def create_ngrams_query_trees(n, trees):
    """
    Forms unique ngram query trees.
    :param n:
    :param trees:
    :return:
    """
    for i in range(n - 1):
        new_trees = []
        for tree in trees:
            # append new_tree only if it is not already inside
            for new_tree in tree_grow(tree):
                duplicate = False
                for confirmed_new_tree in new_trees:
                    if compare_trees(new_tree, confirmed_new_tree):
                        duplicate = True
                        break
                if not duplicate:
                    new_trees.append(new_tree)

        trees = new_trees
    return trees


def split_query_text(input_string):
    """
    Splits query by ignoring everything in brackets and otherwise splitting by spaces.
    :param input_string: Raw query in string
    :return: Split string
    """

    replacements = {}
    brackets_count = 1
    brackets_depth = 0
    replace_string = ''

    for char in input_string:
        if char == '(':
            brackets_depth += 1

        if brackets_depth >= 1:
            replace_string += char

        if char == ')':
            brackets_depth -= 1
            if brackets_depth == 0:
                input_string = input_string.replace(replace_string, f'<BRACKET{brackets_count}>', 1)
                replacements[f'<BRACKET{brackets_count}>'] = replace_string
                brackets_count += 1
                replace_string = ''

    return [el if el not in replacements else replacements[el] for el in input_string.split()]


def decode_query(orig_query, dependency_type):
    """
    Reads query and returns it in tree-form dictionary.
    :param orig_query:
    :param dependency_type:
    :return:
    """
    new_query = False

    # if command in bracelets remove them and treat command as new query
    if orig_query[0] == '(' and orig_query[-1] == ')':
        new_query = True
        orig_query = orig_query[1:-1]

    decoded_query = {'restrictions': []}
    if dependency_type != '':
        dependency_restrictions = []
        for dependency_type_el in dependency_type.split('|'):
            dependency_type_el = dependency_type_el[1:]
            if dependency_type_el and dependency_type_el[0] == '!':
                negation = True
                dependency_type_el = dependency_type_el[1:]
            else:
                negation = False
            if dependency_type_el:
                dependency_restrictions.append((negation, dependency_type_el))
    else:
        dependency_restrictions = None

    if orig_query == '_':
        if dependency_restrictions:
            decoded_query['restrictions'] = [{'deprel': dependency_restriction}
                                             for dependency_restriction in dependency_restrictions]
        return decoded_query
    # if no spaces in query then this is query node and do this otherwise further split query
    elif len(orig_query.split(' ')) == 1:
        for orig_query_or_split_parts in orig_query.split(' ')[0].split('|'):
            restriction = {}
            orig_query_split_parts = orig_query_or_split_parts.split(' ')[0].split('&')
            for orig_query_split_part in orig_query_split_parts:
                if orig_query_split_part[0] == '!' and len(orig_query_split_part) > 1:
                    negation = True
                    orig_query_split_part = orig_query_split_part[1:]
                else:
                    negation = False
                orig_query_split = orig_query_split_part.split('=', 1)
                if len(orig_query_split) > 1:
                    if orig_query_split[0] == 'L':
                        restriction['lemma'] = (negation, orig_query_split[1])
                    elif orig_query_split[0] == 'upos':
                        restriction['upos'] = (negation, orig_query_split[1])
                    elif orig_query_split[0] == 'xpos':
                        restriction['xpos'] = (negation, orig_query_split[1])
                    elif orig_query_split[0] == 'form':
                        restriction['form'] = (negation, orig_query_split[1])
                    elif orig_query_split[0] == 'feats':
                        restriction['feats'] = (negation, orig_query_split[1])
                    elif orig_query_split[0] in UNIVERSAL_FEATURES:
                        restriction['feats_detailed'] = {}
                        restriction['feats_detailed'][orig_query_split[0]] = (negation, orig_query_split[1])
                    elif not new_query:
                        raise Exception('Not supported yet!')
                    else:
                        raise Exception('Unexpected behaviour!')
                elif not new_query:
                    restriction['form'] = (negation, orig_query_split_part)
            decoded_query['restrictions'].append(restriction)

        # merge restrictions from dependencies and other restrictions (this solves 'or' cases in both simultaneously)
        if dependency_restrictions:
            new_restrictions = []
            for dependency_restriction in dependency_restrictions:
                for restriction in decoded_query['restrictions']:
                    new_restriction = restriction.copy()
                    new_restriction['deprel'] = dependency_restriction
                    new_restrictions.append(new_restriction)
            decoded_query['restrictions'] = new_restrictions
        return decoded_query

    # split over spaces if not inside braces
    all_orders = split_query_text(orig_query)

    node_actions = all_orders[::2]
    priority_actions = all_orders[1::2]
    priority_actions_beginnings = [a[0] for a in priority_actions]

    # find root index
    try:
        root_index = priority_actions_beginnings.index('>')
    except ValueError:
        root_index = len(priority_actions)

    children = []
    root = None
    for i, node_action in enumerate(node_actions):
        if i < root_index:
            children.append(decode_query(node_action, priority_actions[i]))
        elif i > root_index:
            children.append(decode_query(node_action, priority_actions[i - 1]))
        else:
            root = decode_query(node_action, dependency_type)
    if children:
        root["children"] = children
    return root


def generate_query_trees(configs, filters):
    """
    Generates query trees based on configs and filters.
    :param configs:
    :param filters:
    :return:
    """
    query_tree = []
    if filters['tree_size_range'][0] > 0:
        if len(filters['tree_size_range']) == 1:
            query_tree = create_ngrams_query_trees(filters['tree_size_range'][0], [{}])
        elif len(filters['tree_size_range']) == 2:
            query_tree = []
            for i in range(filters['tree_size_range'][0], filters['tree_size_range'][1] + 1):
                query_tree.extend(create_ngrams_query_trees(i, [{}]))
    else:
        if filters['tree_size_range'][0] == 0 and 'query' not in configs:
            raise ValueError('You should specify either tree_size or query!')
        query = configs['query']
        query_tree = [decode_query('(' + query + ')', '')]
        if query_tree == [{}]:
            raise ValueError('Query is not formatted properly!')

    return query_tree


def get_query_tree_size(query_tree):
    """
    Returns size of specific query tree.
    :param query_tree:
    :return:
    """
    size = 1
    if 'children' in query_tree:
        for child in query_tree['children']:
            size += get_query_tree_size(child)

    return size


def get_query_tree_size_range(query_trees):
    """
    Returns tree size range.
    :param query_trees:
    :return:
    """
    min_size = 1000000
    max_size = 0
    for query_tree in query_trees:
        size = get_query_tree_size(query_tree)
        if size > max_size:
            max_size = size
        if size < min_size:
            min_size = size
    return [min_size, max_size]



# Copyright 2019 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stark.utils import create_output_string_lemma


class RepresentationNode(object):
    def __init__(self, node, architecture_order, create_output_strings):
        self.name_parts, self.name = self.generate_name(node, create_output_strings)
        self.location = architecture_order
        self.node = node
        for k, v in node.feats_detailed.items():
            assert len(v.keys()) == 1

        self.feats = {k: list(v.keys())[0] for k, v in node.feats_detailed.items()}

    def __repr__(self):
        return self.name

    @staticmethod
    def generate_name(node, create_output_strings, print_lemma=True):
        array = [create_output_string(node) for create_output_string in create_output_strings]
        if create_output_string_lemma in create_output_strings and print_lemma:
            name_array = [create_output_string(
                node) if create_output_string != create_output_string_lemma else create_output_string(node) for
                          create_output_string in create_output_strings]
        else:
            name_array = array
        if len(array) > 1:
            name = '&'.join(name_array)
        else:
            name = name_array[0]

        return array, name



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gc
import random
from abc import abstractmethod
from multiprocessing import Pool
from tqdm import tqdm


class Counter(object):
    """
    A class designed for counting subtrees.
    """
    def __init__(self, document, summary, filters, configs):
        self.document = document
        self.summary = summary
        self.filters = filters
        self.configs = configs

    def run(self):
        """
        Starts counting.
        :return:
        """
        if self.filters['cpu_cores'] > 1:
            self.run_multiprocessor()
        else:
            self.run_single_processor()

    @staticmethod
    @abstractmethod
    def tree_calculations(input_data):
        return []

    def run_multiprocessor(self):
        """
        Runs processing on multiple cores.
        :return:
        """
        with Pool(self.filters['cpu_cores']) as p:
            all_unigrams = p.map(self.get_unigrams,
                                 [(tree, self.filters) for tree in self.document.trees])
            for unigrams in all_unigrams:
                for unigram in unigrams:
                    if unigram in self.summary.unigrams:
                        self.summary.unigrams[unigram] += 1
                    else:
                        self.summary.unigrams[unigram] = 1

            i = 0
            with tqdm(desc='Creating subtrees', total=len(self.document.trees)) as pbar:
                for subtrees in p.imap(self.tree_calculations,
                                                [(tree, self.summary.query_trees, self.filters) for tree in self.document.trees]):

                    for subtree in subtrees:
                        self.postprocess_query_results(subtree, self.document.sentence_statistics[i])
                    i += 1
                    pbar.update()
                    # if i % 1000 == 0:
                    #     gc.collect()

    def run_single_processor(self):
        """
        Runs processing on single core.
        :return:
        """
        for tree, sentence in tqdm(zip(self.document.trees, self.document.sentence_statistics), desc='Processing',
                                   total=len(self.document.trees)):
            input_data = (tree, self.summary.query_trees, self.filters)
            if self.filters['association_measures']:
                unigrams = self.get_unigrams((tree, self.filters))
                for unigram in unigrams:
                    if unigram in self.summary.unigrams:
                        self.summary.unigrams[unigram] += 1
                    else:
                        self.summary.unigrams[unigram] = 1

            subtrees = self.tree_calculations(input_data)
            for subtree in subtrees:
                self.postprocess_query_results(subtree, sentence)

    @staticmethod
    def get_unigrams(input_data):
        """
        Returns unigrams.
        :param input_data:
        :return:
        """
        tree, filters = input_data
        unigrams = []
        # there might be multiple roots in a sentence/tree
        for tree_root in tree:
            unigrams += tree_root.get_unigrams(filters['create_output_string_functs'])
        return unigrams

    def recreate_sentence(self, sentence, r):
        """
        Recreates sentence for example or detailed results.
        :param sentence:
        :param r:
        :return:
        """
        recreated_sentence = ''
        subtree_node_positions = []
        for token_i, token in enumerate(sentence['tokens']):
            subtree_node_positions = r.get_order(self.filters)
            if token_i + 1 in subtree_node_positions:
                letter_position = subtree_node_positions.index(token_i + 1)
                order_letters = r.get_order_letters(subtree_node_positions)
                recreated_sentence += f'{order_letters[letter_position]}[{token[0]}]'
            else:
                recreated_sentence += token[0]
            if token[1]:
                recreated_sentence += ' '
        return recreated_sentence, subtree_node_positions

    def postprocess_query_results(self, r, sentence):
        """
        Gathers processing results, formats and stores them into Summary object.
        :param r:
        :param sentence:
        :return:
        """
        key_raw, word_array = r.get_key_array(self.filters)
        if self.filters['ignored_labels']:
            if self.filters['display_size_range'][0] and \
                    (len(word_array) > self.filters['display_size_range'][-1] or len(word_array) <
                     self.filters['display_size_range'][0]):
                return
        if self.filters['node_order']:
            order_letters = r.get_order_letters(r.get_order(self.filters))
            key = key_raw + order_letters
        else:
            key = key_raw
        sentence_size = len(sentence['tokens']) if 'tokens' in sentence else 10000
        if key in self.summary.representation_trees:
            if self.filters['detailed_results_file']:
                recreated_sentence, subtree_node_positions = self.recreate_sentence(sentence, r)
                sentence_conll = (r.node.node.get_root().conll, subtree_node_positions) if self.filters['annodoc'] else None

                self.summary.representation_trees[key]['sentence'].append((sentence['id'],
                                                                           recreated_sentence,
                                                                           sentence_conll,
                                                                           sentence_size))
            elif self.filters['example'] and self.summary.representation_trees[key]['sentence'][0][
                3] >= 15 and self.summary.representation_trees[key]['sentence'][0][3] > sentence_size:
                recreated_sentence, subtree_node_positions = self.recreate_sentence(sentence, r)
                sentence_conll = (r.node.node.get_root().conll, subtree_node_positions) if self.filters['annodoc'] else None
                self.summary.representation_trees[key]['sentence'] = [(sentence['id'],
                                                                       recreated_sentence,
                                                                       sentence_conll,
                                                                       sentence_size)]
            self.summary.representation_trees[key]['number'] += 1
        else:
            self.summary.representation_trees[key] = {'number': 1, 'key': key_raw, 'word_array': word_array}

            if self.configs['grew_match']:
                self.summary.representation_trees[key]['grew'] = r.get_grew()
                self.summary.representation_trees[key]['location'] = r.get_location_mapper(self.filters)

                # recreate example sentence with shown positions of subtree
            if self.filters['example'] or self.filters['detailed_results_file']:
                recreated_sentence, subtree_node_positions = self.recreate_sentence(sentence, r)
                sentence_conll = (r.node.node.get_root().conll, subtree_node_positions) if self.filters['annodoc'] else None
                sentence_size = len(sentence['tokens']) if 'tokens' in sentence else 10000
                self.summary.representation_trees[key]['sentence'] = [(sentence['id'],
                                                                       recreated_sentence,
                                                                       sentence_conll,
                                                                       sentence_size)]
            if self.filters['node_order']:
                self.summary.representation_trees[key]['order_letters'] = order_letters
                if self.configs['depsearch']:
                    self.summary.representation_trees[key]['key_sorted'] = r.get_key_sorted(self.filters)[1:-1]
            if self.filters['head_info']:
                self.summary.representation_trees[key]['root_name'] = r.node.name
            if self.configs['greedy_counter'] and self.summary.max_tree_size < r.tree_size:
                self.summary.max_tree_size = r.tree_size

        if self.filters['sentence_count_file']:
            if key in sentence['count']:
                sentence['count'][key] += 1
            else:
                sentence['count'][key] = 1


class QueryCounter(Counter):
    """
    Counter that generates only trees that fit query trees.
    """
    def __init__(self, *configs):
        super().__init__(*configs)

    @staticmethod
    def tree_calculations(input_data):
        tree, query_trees, filters = input_data
        subtrees = []
        # there might be multiple roots in a sentence/tree
        for tree_root in tree:
            _, subtrees_part = tree_root.get_subtrees(query_trees, [], filters)
            subtrees += subtrees_part
        return [subtree for query_results in subtrees for subtree in query_results]


class GreedyCounter(Counter):
    """
    Counter that generates all trees and then filters them.
    """

    def __init__(self, *configs):
        super().__init__(*configs)

    @staticmethod
    def tree_calculations(input_data):
        tree, query_trees, filters = input_data
        subtrees = []
        # there might be multiple roots in a sentence/tree
        for tree_root in tree:
            _, subtrees_part = tree_root.get_subtrees(filters)
            subtrees += subtrees_part

        subtrees = GreedyCounter.filter_subtrees(query_trees, subtrees, filters)
        return subtrees

    @staticmethod
    def filter_subtrees(query_trees, subtrees, filters):
        """
        Filters subtrees and ignores labels if necessary.
        :param query_trees:
        :param subtrees:
        :param filters:
        :return:
        """
        return [subtree.ignore_labels(filters) for subtree in subtrees if subtree.pass_filter(query_trees, filters)]



import filecmp
import os
import random
import shutil

import pytest
import stark
from stark.stark import read_settings, parse_args
from tests import *


def test_base():
    """
    Testing regular run.
    :return:
    """
    random.seed(12)
    if not os.path.exists(OUTPUT_DIR):
        os.mkdir(OUTPUT_DIR)
    config_file = os.path.join(CONFIGS_DIR, 'config_base.ini')
    settings = read_settings(config_file, parse_args([]))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_base.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_base.tsv'))


def test_greedy_complete():
    """
    Testing regular run.
    :return:
    """
    random.seed(12)
    if not os.path.exists(OUTPUT_DIR):
        os.mkdir(OUTPUT_DIR)
    config_file = os.path.join(CONFIGS_DIR, 'config_greedy_complete.ini')
    settings = read_settings(config_file, parse_args([]))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_greedy_complete.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_greedy_complete.tsv'))


def test_compare():
    """
    Test compare.
    :return:
    """
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_compare.ini')
    settings = read_settings(config_file, parse_args([]))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_compare.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_compare.tsv'))
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_compare.ini')
    settings = read_settings(config_file, parse_args(['--greedy_counter', 'yes']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_compare.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_compare.tsv'))


def test_query():
    """
    Test complete=no and query.
    :return:
    """
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_query.ini')
    settings = read_settings(config_file, parse_args([]))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_query.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_query.tsv'))

    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_query.ini')
    settings = read_settings(config_file, parse_args(['--greedy_counter', 'yes']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_query.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_query.tsv'))


def test_fixed():
    """
    Test complete=no and query.
    :return:
    """
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_base.ini')
    settings = read_settings(config_file, parse_args(['--fixed', 'no',
                                                      '--output', 'test_data/output/out_fixed.tsv']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_fixed.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_fixed.tsv'))

    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_base.ini')
    settings = read_settings(config_file, parse_args(['--greedy_counter', 'yes',
                                                      '--fixed', 'no',
                                                      '--output', 'test_data/output/out_fixed.tsv']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_fixed.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_fixed.tsv'))


def test_dir():
    """
    Test complete=no and query.
    :return:
    """
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_base.ini')
    settings = read_settings(config_file, parse_args(['--input', 'test_data/input/dir_input/',
                                                      '--output', 'test_data/output/out_dir.tsv']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_dir.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_dir.tsv'))
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_base.ini')
    settings = read_settings(config_file, parse_args(['--input', 'test_data/input/dir_input/',
                                                      '--output', 'test_data/output/out_dir.tsv',
                                                      '--greedy_counter', 'yes']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_dir.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_dir.tsv'))


def test_internal_storage():
    """
    Test internal storage.
    :return:
    """
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_internal_storage.ini')
    output_mapper_dir = 'test_data/output/internal_saves'
    if os.path.exists(output_mapper_dir):
        shutil.rmtree(output_mapper_dir)
    settings = read_settings(config_file, parse_args([]))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_dir.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_dir.tsv'))

    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_internal_storage.ini')
    output_mapper_dir = 'test_data/output/internal_saves'
    if os.path.exists(output_mapper_dir):
        shutil.rmtree(output_mapper_dir)
    settings = read_settings(config_file, parse_args(['--greedy_counter', 'yes']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_dir.tsv'), os.path.join(CORRECT_OUTPUT_DIR, 'out_dir.tsv'))


def test_internal_storage2():
    """
    Test internal storage.
    :return:
    """
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_base.ini')
    output_mapper_dir = 'test_data/output/internal_saves'
    if os.path.exists(output_mapper_dir):
        shutil.rmtree(output_mapper_dir)
    settings = read_settings(config_file, parse_args(['--internal_saves', 'test_data/output/internal_saves',
                                                      '--output', 'test_data/output/out_internal_storage2.tsv']))
    stark.run(settings)
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_internal_storage2.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                           'out_internal_storage2.tsv'))

    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_base.ini')
    output_mapper_dir = 'test_data/output/internal_saves'
    if os.path.exists(output_mapper_dir):
        shutil.rmtree(output_mapper_dir)
    settings = read_settings(config_file, parse_args(['--internal_saves', 'test_data/output/internal_saves',
                                                      '--output', 'test_data/output/out_internal_storage2.tsv',
                                                      '--greedy_counter', 'yes']))
    stark.run(settings)
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_internal_storage2.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                           'out_internal_storage2.tsv'))


def test_output_settings():
    """
    Test complete=no and query.
    :return:
    """
    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_output_settings.ini')
    settings = read_settings(config_file, parse_args(['--detailed_results_file', 'test_data/output/detailed_results_file_query.tsv']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_output_settings.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                         'out_output_settings.tsv'))
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'detailed_results_file_query.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                           'detailed_results_file_query.tsv'))
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'sentence_count_file.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                         'sentence_count_file.tsv'))

    random.seed(12)
    config_file = os.path.join(CONFIGS_DIR, 'config_output_settings.ini')
    settings = read_settings(config_file, parse_args(['--detailed_results_file',
                                                      'test_data/output/detailed_results_file_greedy.tsv',
                                                      '--sentence_count_file',
                                                      'test_data/output/sentence_count_file_greedy.tsv',
                                                      '--greedy_counter', 'yes']))
    stark.run(settings)
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'out_output_settings.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                         'out_output_settings.tsv'))
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'detailed_results_file_greedy.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                           'detailed_results_file_greedy.tsv'))
    assert filecmp.cmp(os.path.join(OUTPUT_DIR, 'sentence_count_file_greedy.tsv'), os.path.join(CORRECT_OUTPUT_DIR,
                                                                                         'sentence_count_file_greedy.tsv'))



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import string
import sys

import networkx as nx
from networkx.algorithms.bipartite import hopcroft_karp_matching

from stark.data.representation.tree import RepresentationTree
from stark.processing.filters import Filter


class GreedyRepresentationTree(RepresentationTree):
    def __init__(self, node, children, filters):
        self.tree_size = children[0]
        if filters['node_order']:
            children_sorted = children[1]
        else:
            children_sorted = sorted(children[1], key=lambda x: (x.get_key(filters), x.node.node.deprel))
        super().__init__(node, children_sorted)

    @classmethod
    def copy(cls, node, children, filters):
        """
        Copies original object with empty children.
        :param filters:
        :param children:
        :param node:
        :return:
        """
        tree_size = sum([child.tree_size for child in children]) + 1
        return cls(node, [tree_size, children], filters)

    def check_query(self, query, filters):
        # compares query and children lengths
        query_length = len(query['children']) if 'children' in query else 0
        if query_length != len(self.children):
            return False

        # does node comparisons
        filt = Filter.check_query_tree(query, self.node.node.form, self.node.node.lemma, self.node.node.upos, self.node.node.xpos,
                                       self.node.node.feats, self.node.node.deprel, self.children, filters)

        if not filt:
            return False

        if not self.children:
            return True

        # compares children with query nodes
        # creates nodes for graph children nodes go from 1 on and query children go from -1 back.
        children_nodes = list(range(1, query_length+1))
        query_children_nodes = list(range(-query_length, 0))
        edges = []
        for i, child in enumerate(self.children):
            for j, query_child in enumerate(query['children']):
                if child.check_query(query_child, filters):
                    edges.append((i + 1, -j - 1))

        graph = nx.Graph()
        graph.add_nodes_from(children_nodes, bipartite=1)
        graph.add_nodes_from(query_children_nodes, bipartite=0)
        graph.add_edges_from(edges)

        # hopcroft-karp algorithm efficiently finds number of connections between
        connections = hopcroft_karp_matching(graph, children_nodes)
        return len(connections) == query_length * 2

    def pass_filter(self, query_trees, filters):
        """
        Validator of filters for greedy counter.
        :return: True when representation tree passes all filters.
        """
        # tests tree size
        if query_trees:
            pass_query = False
            for query in query_trees:
                if self.check_query(query, filters):
                    pass_query = True
            if not pass_query:
                return False

        # drop too small trees
        if (filters['display_size_range'][-1] and not (filters['display_size_range'][0]
                <= self.tree_size <= filters['display_size_range'][-1])):
            return False

        return Filter.check_representation_tree(self, filters)

    # def get_size(self):
    #     size = 0
    #     size += sys.getsizeof(self.node)
    #     size += sys.getsizeof(self.tree_size)
    #     size += sys.getsizeof(self.children)
    #     for child in self.children:
    #         size += child.get_size()
    #     return size


# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import pickle
import re
from pathlib import Path


def save_zipped_pickle(obj, filename, protocol=-1):
    Path(filename).parent.mkdir(parents=True, exist_ok=True)
    with gzip.open(filename, 'wb') as f:
        pickle.dump(obj, f, protocol)


def load_zipped_pickle(filename):
    with gzip.open(filename, 'rb') as f:
        loaded_object = pickle.load(f)
        return loaded_object


def printable_answers(query):
    all_orders = re.split(r"\s+(?=[^()]*(?:\(|$))", query)
    node_actions = all_orders[::2]

    if len(node_actions) > 1:
        res = []
        for node_action in node_actions:
            # if command in bracelets remove them and treat command as new query
            if not node_action:
                res.extend(['('])
            elif node_action[0] == '(' and node_action[-1] == ')':
                res.extend(printable_answers(node_action[1:-1]))
            else:
                res.extend([node_action])
        return res
    else:
        return [query]


def create_output_string_form(tree):
    return tree.form


def create_output_string_deprel(tree):
    return tree.deprel


def create_output_string_lemma(tree):
    return tree.lemma if tree.lemma is not None else '_'


def create_output_string_upos(tree):
    return tree.upos


def create_output_string_xpos(tree):
    return tree.xpos


def create_output_string_none(tree):
    return '_'


def create_output_string_feats(tree):
    return '|'.join([f'{k}={list(v.keys())[0]}' for k, v in tree.feats_detailed.items()])


def create_output_string_misc(tree):
    return tree.misc



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Document(object):
    def __init__(self):
        self.trees = []
        self.sentence_statistics = []
        self.form_dict = {}
        self.lemma_dict = {}
        self.upos_dict = {}
        self.xpos_dict = {}
        self.deprel_dict = {}
        self.misc_dict = {}

    def get_document_data(self):
        return [self.trees, self.form_dict, self.lemma_dict, self.upos_dict, self.xpos_dict, self.deprel_dict, self.misc_dict,
                self.sentence_statistics]

    @classmethod
    def create_document_from_cache(cls, doc_data):
        d = cls()
        d.trees, d.form_dict, d.lemma_dict, d.upos_dict, d.xpos_dict, d.deprel_dict, d.misc_dict, d.sentence_statistics = doc_data
        return d



#!/usr/bin/env python
# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import hashlib
import json
import math
import os
import random
import shutil
import string
from abc import abstractmethod
from pathlib import Path
import sys
from os import path
import urllib.parse
import logging
from tqdm import tqdm

here = path.abspath(path.dirname(__file__))
logging.basicConfig(level=logging.NOTSET)
logger = logging.getLogger('stark')


class Writer(object):
    """
    A class used for writing results.
    """
    def __init__(self, summary, other_summary, filters, configs):
        self.summary = summary
        self.other_summary = other_summary
        self.filters = filters
        self.configs = configs

    @abstractmethod
    def write(self):
        return

    def lines_generator(self):
        """
        A generator that returns lines in array form, that can be used for further processing.
        :return:
        """
        other_representation_trees = self.other_summary.representation_trees if self.other_summary else None
        other_corpus_size = self.other_summary.corpus_size if self.other_summary else None
        random_sentence_position = 0

        # skip elements that do not fit filters
        if self.filters['frequency_threshold'] or self.filters['display_size_range'][-1]:
            # reset summary_size because some elements will be skipped!
            self.summary.max_tree_size = 1
            sorted_list = []
            for k, v in self.summary.representation_trees.items():
                # skip words that do not fit display size (when it is given)
                if (self.filters['display_size_range'][-1] and not
                        self.filters['display_size_range'][0] <= len(v['word_array']) <=
                        self.filters['display_size_range'][-1]):
                    continue

                if self.filters['frequency_threshold'] and self.filters['frequency_threshold'] > v['number']:
                    continue
                if self.summary.max_tree_size < len(v['word_array']):
                    self.summary.max_tree_size = len(v['word_array'])
                sorted_list.append((k, v))
        else:
            sorted_list = self.summary.representation_trees.items()

        sorted_list = sorted(sorted_list, key=lambda x: (-x[1]['number'], x[0]))

        with open(os.path.join(here, '../resources/codes_mapper.json'), 'r') as f:
            codes_mapper = json.load(f)
        path = Path(self.configs['input_path']).name
        lang = path.split('_')[0]
        corpus_name = path.split('_')[1].split('-')[0].lower() if len(path.split('_')) > 1 else 'unknown'
        corpus = codes_mapper[lang][corpus_name] if lang in codes_mapper and corpus_name in codes_mapper[lang] else None

        if self.configs['sentence_count_file']:
            self.write_sentence_count_file()

        if self.configs['detailed_results_file']:
            self.write_detailed_results_file()

        if self.configs['annodoc_example_dir'] and (self.configs['detailed_results_file'] or self.configs['example']):
            self.write_annodoc_files()

        if self.configs['annodoc_detailed_dir'] and self.configs['detailed_results_file']:
            self.write_annodoc_detailed_files()

        if self.filters['display_size_range'][-1]:
            if not self.configs['greedy_counter']:
                len_words = self.filters['display_size_range'][-1]
            else:
                len_words = min(self.filters['display_size_range'][-1], self.summary.max_tree_size)
        else:
            len_words = int(len(self.configs['query'].split(" ")) / 2 + 1)
        header = ["Tree"]
        if self.configs['node_info']:
            header += ["Node " + string.ascii_uppercase[i % 26] + str(int(i/26)) + "-" + node_type if i >= 26 else
                             "Node " + string.ascii_uppercase[i % 26] + "-" + node_type for i in range(len_words) for
                             node_type in self.filters['node_types']]
        header += ['Absolute frequency', 'Relative frequency']
        if self.filters['node_order']:
            header += ['Order']
        if self.configs['grew_match']:
            header += ['Grew-match query']
        if corpus and self.configs['grew_match']:
            header += ['Grew-match URL']
        if self.filters['node_order'] and self.configs['depsearch']:
            header += ['Depsearch tree query']
        if self.filters['nodes_number']:
            header += ['Number of nodes']
        if self.filters['head_info']:
            header += ['Head node']
        if self.filters['example']:
            header += ['Example']
        if self.filters['annodoc']:
            header += ['Annodoc']
        if self.filters['association_measures']:
            header += ['MI', 'MI3', 'Dice', 'logDice', 't-score', 'simple-LL']
        if self.configs['compare']:
            header += ['Absolute frequency in second treebank', 'Relative frequency in second treebank', 'Ratio', 'LL',
                       'BIC', 'Log ratio', 'OR', '%DIFF']
        yield header

        if self.filters['lines_threshold']:
            sorted_list = sorted_list[:self.filters['lines_threshold']]

        # body
        for k, v in tqdm(sorted_list, desc='Writing'):
            literal_key = v['key']
            word_array = v['word_array']

            relative_frequency = v['number'] * 1000000.0 / self.summary.corpus_size
            words_only = [word_att for word in word_array for word_att in word] + ['' for _ in range(
                (len_words - len(word_array)) * len(word_array[0]))]
            key = literal_key[1:-1] if (len(literal_key) > 0 and literal_key[0] == '(' and literal_key[-1] == ')') else literal_key

            row = [key]
            if self.configs['node_info']:
                row += words_only
            row += [str(v['number']), '%.1f' % relative_frequency]
            if self.filters['node_order']:
                order_letters = v['order_letters']
                row += [order_letters]
            if self.configs['grew_match']:
                location_mapper = v['location']
                grew_nodes, grew_links = v['grew']
                key_grew = self.get_grew(grew_nodes, grew_links, self.filters['node_types'], self.filters['node_order'],
                                         location_mapper, self.filters['dependency_type'],
                                         self.filters['complete_tree_type'])
                row += [key_grew]
                if corpus:
                    url = f'http://universal.grew.fr/?corpus={corpus}&request={urllib.parse.quote(key_grew)}'
                    row += [url]
            if self.filters['node_order'] and self.configs['depsearch']:
                row += [v['key_sorted']]
            if self.filters['nodes_number']:
                row += ['%d' % len(word_array)]
            if self.filters['head_info']:
                row += [v['root_name']]
            if self.filters['example']:
                random_sentence_position = 0
                min_sentence_size = 100000
                final_row = [v['sentence'][random_sentence_position][1]]
                for i, s in enumerate(v['sentence']):
                    if s[3] < 15:
                        final_row = [s[1]]
                        random_sentence_position = i
                        break
                    elif s[3] < min_sentence_size:
                        final_row = [s[1]]
                        random_sentence_position = i
                        min_sentence_size = s[3]
                row += final_row
            if self.filters['annodoc'] and (self.configs['detailed_results_file'] or self.filters['example']):
                annodoc_dict = {'id': v['sentence'][random_sentence_position][0],'positions': v['sentence'][random_sentence_position][2][1],'subtree_hash': hashlib.sha1(k.encode('utf-8')).hexdigest()}
                annodoc_json = json.dumps(annodoc_dict)
                row += [annodoc_json]
            if self.filters['association_measures']:
                row += self.get_collocabilities(v, self.summary.unigrams, self.summary.corpus_size)
            if self.configs['compare']:
                other_abs_freq = other_representation_trees[k]['number'] if k in other_representation_trees else 0
                row += self.get_keyness(v['number'], other_abs_freq, self.summary.corpus_size, other_corpus_size)
            yield row

    def write_sentence_count_file(self):
        """
        Writes into sentence count file.
        :return:
        """
        if os.path.exists(self.configs['sentence_count_file']):
            os.remove(self.configs['sentence_count_file'])
        with open(self.configs['sentence_count_file'], "a", newline="", encoding="utf-8") as wf:
            key_list = [k for k, v in self.summary.representation_trees.items()]
            header = ['Sentence_id'] + key_list
            wf.write('\t'.join(header) + '\n')
            for sentence in self.summary.samples:
                wf.write(sentence['id'] + '\t' + '\t'.join(
                    [str(sentence['count'][k]) if k in sentence['count'] else '0' for k in key_list]) + '\n')

    def write_detailed_results_file(self):
        """
        Writes into detailed results file.
        :return:
        """
        if os.path.exists(self.configs['detailed_results_file']):
            os.remove(self.configs['detailed_results_file'])
        with open(self.configs['detailed_results_file'], "a", newline="", encoding="utf-8") as wf:
            for k, v in self.summary.representation_trees.items():
                for s in v['sentence']:
                    wf.write(k + '\t' + s[0] + '\t' + s[1] + '\n')

    def write_annodoc_files(self):
        """
        Writes conllu files separated by examples.
        :return:
        """
        annodoc_dir = Path(self.configs['annodoc_example_dir'])
        if annodoc_dir.exists():
            shutil.rmtree(annodoc_dir, ignore_errors=True)
        annodoc_dir.mkdir()
        for k, v in self.summary.representation_trees.items():
            for s in v['sentence']:
                annodoc_path = Path(self.configs['annodoc_example_dir'], s[0])
                if not annodoc_path.exists():
                    with open(annodoc_path, "w", newline="",
                              encoding="utf-8") as wf:
                        wf.write(s[2][0])

    def write_annodoc_detailed_files(self):
        """
        Writes tsv files that contain hashes of subtrees in a name. Each file contains a list with sentence_id and
        positions of subtree in a sentence.
        :return:
        """
        annodoc_dir = Path(self.configs['annodoc_detailed_dir'])
        if annodoc_dir.exists():
            shutil.rmtree(annodoc_dir, ignore_errors=True)
        annodoc_dir.mkdir()
        for k, v in self.summary.representation_trees.items():
            path = hashlib.sha1(k.encode('utf-8')).hexdigest()
            annodoc_path = Path(self.configs['annodoc_detailed_dir'], path) # calculate hash?
            if not annodoc_path.exists():
                with open(annodoc_path, "w", newline="",
                          encoding="utf-8") as wf:
                    for s in v['sentence']:
                        wf.write(f'{str(s[0])}\t{str(s[2][1])}\n')

    @staticmethod
    def get_keyness(abs_freq_A, abs_freq_B, count_A, count_B):
        """
        Calculates keyness for statistic purposes.
        :param abs_freq_A:
        :param abs_freq_B:
        :param count_A:
        :param count_B:
        :return:
        """
        ratio = '%.2f' % ((abs_freq_A / count_A) / (abs_freq_B / count_B)) if abs_freq_B else 'NaN'
        if abs_freq_B <= 0:
            abs_freq_B = 0.000000000000000001
        E1 = count_A * (abs_freq_A + abs_freq_B) / (count_A + count_B)
        E2 = count_B * (abs_freq_A + abs_freq_B) / (count_A + count_B)

        LL = 2 * ((abs_freq_A * math.log(abs_freq_A / E1)) + (
                abs_freq_B * math.log(abs_freq_B / E2))) if abs_freq_B > 0 else 0
        BIC = LL - math.log(count_A + count_B) if abs_freq_B > 0 else 0
        log_ratio = math.log(((abs_freq_A / count_A) / (abs_freq_B / count_B)), 2) if abs_freq_B > 0 else 0
        if count_A == abs_freq_A or count_B == abs_freq_B:
            OR = 'NaN'
        else:
            OR = (abs_freq_A / (count_A - abs_freq_A)) / (abs_freq_B / (count_B - abs_freq_B)) if abs_freq_B > 0 else 0
            OR = '%.2f' % OR
        diff = (((abs_freq_A / count_A) * 1000000 - (abs_freq_B / count_B) * 1000000) * 100) / (
                (abs_freq_B / count_B) * 1000000) if abs_freq_B > 0 else 0

        if abs_freq_B <= 0:
            return ['%.0f' % abs_freq_B, '%.1f' % (abs_freq_B * 1000000.0 / count_B), ratio, LL, BIC, log_ratio, OR,
                    diff]
        return ['%.0f' % abs_freq_B, '%.1f' % (abs_freq_B * 1000000.0 / count_B), ratio, '%.2f' % LL,
                '%.2f' % BIC, '%.2f' % log_ratio, OR, '%.2f' % diff]

    @staticmethod
    def get_grew(nodes, links, node_types, node_order, location_mapper, dependency_type, complete):
        """
        Forms grew format for usage and comparison on external website.
        :param nodes:
        :param links:
        :param node_types:
        :param node_order:
        :param location_mapper:
        :param dependency_type:
        :param complete:
        :return:
        """
        assert nodes
        node_result = {}
        for node in nodes:
            node_result[location_mapper[node.location]] = []
            for node_type in node_types:
                if node_type == 'deprel':
                    node_result[location_mapper[node.location]].append(f'deprel={node.node.deprel}')
                elif node_type == 'lemma':
                    node_result[location_mapper[node.location]].append(f'lemma="{node.node.lemma}"')
                elif node_type == 'upos':
                    node_result[location_mapper[node.location]].append(f'upos={node.node.upos}')
                elif node_type == 'xpos':
                    node_result[location_mapper[node.location]].append(f'xpos={node.node.xpos}')
                elif node_type == 'generic':
                    node_result[location_mapper[node.location]].append(f'')
                elif node_type == 'feats':
                    for k, v in node.feats.items():
                        node_result[location_mapper[node.location]].append(f'{k}={v}')
                else:
                    node_result[location_mapper[node.location]].append(f'form="{node.node.form}"')
        link_result = []
        order_result = []
        for link in links:
            link_result.append([location_mapper[link[0].location], location_mapper[link[1].location]])
            if node_order:
                if link[0].location < link[1].location:
                    order_result.append([location_mapper[link[0].location], location_mapper[link[1].location], '<<'])
                else:
                    order_result.append([location_mapper[link[0].location], location_mapper[link[1].location], '>>'])
        grew = 'pattern {'
        grew += '; '.join([node_k + ' [' + ', '.join(v) + ']' for node_k, v in node_result.items()])
        if links:
            grew += '; '
            grew += '; '.join([f'{link[0]} -[{link_node[1].node.deprel}]-> {link[1]}' for link, link_node in
                               zip(link_result, links)]) if dependency_type else \
                '; '.join([f'{link[0]} -> {link[1]}' for link in link_result])
            grew += '; ' + '; '.join([f'{link[0]} {link[2]} {link[1]}' for link in order_result])

        if complete:
            without_statements = []
            for i, key in enumerate(node_result.keys()):
                without_statements.append(f'without {{{key} -> X{i}}}')

            return grew + '} ' + ' '.join(without_statements)
        return grew + '}'

    @staticmethod
    def get_collocabilities(ngram, unigrams, corpus_size):
        """
        Calculates collocabilities.
        :param ngram:
        :param unigrams:
        :param corpus_size:
        :return:
        """
        # n of ngram
        n = len(ngram['word_array'])

        # collocabilities are supported for n <= 10
        if n > 10:
            return ['NaN'] * 6

        sum_fwi = 0.0
        mul_fwi = 1.0
        for key_array in ngram['word_array']:
            # create key for unigrams
            if len(key_array) > 1:
                key = '&'.join(key_array)
            else:
                key = key_array[0]
            sum_fwi += unigrams[key]
            mul_fwi *= unigrams[key]

        if mul_fwi < 0:
            mul_fwi = sys.maxsize

        # number of all words
        N = corpus_size

        O = ngram['number']
        E = mul_fwi / pow(N, n - 1)

        # ['MI', 'MI3', 'Dice', 'logDice', 't-score', 'simple-LL']
        mi = math.log(O / E, 2)
        mi3 = math.log(pow(O, 3) / E, 2)
        dice = n * O / sum_fwi
        logdice = 14 + math.log(dice, 2)
        tscore = (O - E) / math.sqrt(O)
        simplell = 2 * (O * math.log10(O / E) - (O - E))
        return ['%.2f' % mi, '%.2f' % mi3, '%.2f' % dice, '%.2f' % logdice, '%.2f' % tscore, '%.2f' % simplell]


class TSVWriter(Writer):
    """
    A class that writes into TSV file.
    """
    def __init__(self, *configs):
        super().__init__(*configs)

    def write(self):
        with open(self.configs['output'], "w", newline="", encoding="utf-8") as f:
            # header - use every second space as a split
            writer = csv.writer(f, delimiter='\t')
            for line in self.lines_generator():
                writer.writerow(line)


class ObjectWriter(Writer):
    """
        A class that returns results in object.
        """
    def __init__(self, *configs):
        super().__init__(*configs)

    def write(self):
        return list(self.lines_generator())



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time
from pathlib import Path

from stark.processing.cache import ProcessorCache
from stark.processing.counters import QueryCounter, GreedyCounter
from stark.processing.document_processor import DocumentProcessor

logger = logging.getLogger('stark')


class Processor(object):
    """
    A class that iterates over all documents and stores cumulative results.
    """
    def __init__(self, configs, filters):
        self.configs = configs
        self.filters = filters

    def run_dir(self, summary):
        """
        Runs processing for directories. When enabled, use caching.
        :param summary:
        :return:
        """
        processor_cache = ProcessorCache(self)
        summary = processor_cache.load_cache(summary)

        for path in sorted(Path(self.configs['input_path']).rglob('*.conllu')):
            summary = processor_cache.process_trees(path, summary)

        return summary

    def run(self, path, summary):
        """
        Run processing.
        :param path_list: List of paths to documents that need to be processed.
        :param summary: A collection of datapoints used for result generation.
        :return:
        summary: A collection of datapoints used for result generation.
        """
        start_exe_time = time.time()

        document_processor = DocumentProcessor(str(path), self)
        document = document_processor.form_trees(summary, self.configs)
        logger.info("Trees formed time:")
        logger.info("--- %s seconds ---" % (time.time() - start_exe_time))
        if self.configs['greedy_counter']:
            tree_counter = GreedyCounter(document, summary, self.filters, self.configs)
        else:
            tree_counter = QueryCounter(document, summary, self.filters, self.configs)
        tree_counter.run()
        summary.samples.extend(document.sentence_statistics)

        logger.info(f"{len(summary.representation_trees)} unique trees counted time (execution time):")
        logger.info("Trees counted time (execution time):")
        logger.info("--- %s seconds ---" % (time.time() - start_exe_time))

        return summary



#!/usr/bin/env python
# Copyright 2019 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import configparser
import os
from pathlib import Path
import sys
import logging

from stark.data.representation.empty_namespace import EmptyNamespace
# from pympler import asizeof

from stark.data.summary import Summary
from stark.processing.filters import read_filters
from stark.processing.processor import Processor
from stark.processing.query_trees import generate_query_trees, get_query_tree_size_range
from stark.processing.writers import TSVWriter, ObjectWriter

logging.basicConfig(level=logging.NOTSET)
logger = logging.getLogger('stark')

sys.setrecursionlimit(25000)


def parse_args(args):
    """
    Returns Namespace object containing given arguments.
    :param args:
    :return:
    """
    parser = argparse.ArgumentParser()

    # Required parameters
    parser.add_argument("--config_file", default=str(Path.joinpath(Path(__file__).parent.parent, "config.ini")), type=str,
                        help="The input config file.")
    parser.add_argument("--input", default=None, type=str, help="The input file/folder.")
    parser.add_argument("--output", default=None, type=str, help="The output file.")
    parser.add_argument("--internal_saves", default=None, type=str, help="Location for internal_saves.")
    parser.add_argument("--cpu_cores", default=None, type=int, help="Number of cores used.")
    parser.add_argument("--greedy_counter", default=None, type=str, help="Uses greedy counter.")

    parser.add_argument("--size", default=None, type=str, help="Size of trees displayed.")
    parser.add_argument("--processing_size", default=None, type=str, help="Limits the size of trees during processing.")
    parser.add_argument("--complete", default=None, type=str, help="Tree type.")
    parser.add_argument("--labeled", default=None, type=str, help="Dependency type.")
    parser.add_argument("--fixed", default=None, type=str, help="Order of node.")
    parser.add_argument("--node_type", default=None, type=str, help="Type of node.")

    parser.add_argument("--allowed_labels", default=None, type=str, help="Label not included here will not be counted.")
    parser.add_argument("--ignored_labels", default=None, type=str,
                        help="Labels that contain these parsing tags will be counted"
                             " but ignored in result representations.")
    parser.add_argument("--head", default=None, type=str, help="Head whitelist.")

    parser.add_argument("--query", default=None, type=str, help="Query.")

    # output settings
    parser.add_argument("--grew_match", default=None, type=str, help="Output setting for printing query and url.")
    parser.add_argument("--label_subtypes", default=None, type=str, help="Output setting for printing query and url.")
    parser.add_argument("--depsearch", default=None, type=str, help="Output setting for printing depsearch query.")
    parser.add_argument("--example", default=None, type=str, help="Print one example in a separate column.")
    parser.add_argument("--sentence_count_file", default=None, type=str,
                        help="Path to a file where counts of queries will be stored per sentence.")
    parser.add_argument("--detailed_results_file", default=None, type=str,
                        help="Path to a file where all examples will be stored.")
    parser.add_argument("--annodoc_example_dir", default=None, type=str,
                        help="Path to a directory where annodoc example files are stored (js library for visualization).")
    parser.add_argument("--annodoc_detailed_dir", default=None, type=str,
                        help="Path to a directory where annodoc detailed files are stored (js library for visualization).")

    parser.add_argument("--max_lines", default=None, type=int, help="Maximum number of trees in the output.")
    parser.add_argument("--node_info", default=None, type=str, help="Information about nodes in separate columns.")
    parser.add_argument("--head_info", default=None, type=str, help="Information about head in separate columns.")
    parser.add_argument("--frequency_threshold", default=None, type=int, help="Frequency threshold.")
    parser.add_argument("--association_measures", default=None, type=str, help="Association measures.")
    parser.add_argument("--continuation_processing", default=None, type=str, help="Nodes number.")
    parser.add_argument("--compare", default=None, type=str, help="Corpus with which we want to compare statistics.")
    return parser.parse_args(args)


def count_subtrees(configs, filters):
    """
    Counts subtrees that match filters.
    :param configs:
    :param filters:
    :return:
    """
    processor = Processor(configs, filters)
    summary = Summary()
    if not configs['greedy_counter'] or filters['tree_size_range'][0] == 0:
        summary.set_query_trees(generate_query_trees(configs, filters))

        # set min and max tree size to limit greedy generation
        if configs['greedy_counter']:
            filters['tree_size_range'] = get_query_tree_size_range(summary.query_trees)

    if os.path.isdir(configs['input_path']):
        summary = processor.run_dir(summary)

    else:
        summary = processor.run(configs['input_path'], summary)

    return summary


def read_configs(config, args=EmptyNamespace()):
    """
    Merges concrete settings from config files with arguments. When arguments are given, they override settings from
    configuration file. Look into documentation for which parameters in config file are required.
    :param config: ConfigParser object containing settings from configuration file.
    :param args: Namespace object containing given arguments.
    :return: Dictionary with configuration settings.
    """

    configs = {}
    # mandatory parameters
    configs['input_path'] = config.get('settings', 'input') if not args.input else args.input
    if config.has_option('settings', 'output') or args.output:
        configs['output'] = config.get('settings', 'output') if not args.output else args.output
    else:
        configs['output'] = None
    configs['display_size'] = config.get('settings', 'size', fallback='0') if not args.size else args.size
    configs['tree_size'] = config.get('settings', 'processing_size', fallback='0') if not args.processing_size else (
        args.processing_size)
    if config.has_option('settings', 'node_type') or args.node_type:
        configs['node_type'] = config.get('settings', 'node_type') if not args.node_type else args.node_type
    else:
        configs['node_type'] = None

    # mandatory parameters with default value
    configs['greedy_counter'] = (config.getboolean('settings', 'greedy_counter') if not args.greedy_counter
                                  else args.greedy_counter == 'yes')
    configs['internal_saves'] = (config.get('settings', 'internal_saves')
                                 if config.has_option('settings', 'internal_saves') else None) \
        if not args.internal_saves else args.internal_saves
    configs['cpu_cores'] = (config.getint('settings', 'cpu_cores') if config.has_option('settings', 'cpu_cores')
                            else 1) if not args.cpu_cores else args.cpu_cores
    configs['complete_tree_type'] = (config.getboolean('settings', 'complete') if not args.complete
                                     else args.complete == 'yes')
    configs['dependency_type'] = (config.getboolean('settings', 'labeled') if not args.labeled
                                  else args.labeled == 'yes')
    configs['node_order'] = (config.getboolean('settings', 'fixed') if not args.fixed else args.fixed == 'yes')
    configs['association_measures'] = (config.getboolean('settings', 'association_measures')
                                       if not args.association_measures else args.association_measures == 'yes')
    configs['node_info'] = (config.getboolean('settings', 'node_info')
                                       if not args.node_info else args.node_info == 'yes')
    configs['head_info'] = (config.getboolean('settings', 'head_info')
                            if not args.head_info else args.head_info == 'yes')

    # optional parameters
    if config.has_option('settings', 'allowed_labels') or args.allowed_labels:
        label_whitelist = config.get('settings',
                                     'allowed_labels') if not args.allowed_labels else args.allowed_labels
        configs['label_whitelist'] = label_whitelist.split('|')
    else:
        configs['label_whitelist'] = []

    if config.has_option('settings', 'ignored_labels') or args.ignored_labels:
        ignored_labels = config.get('settings', 'ignored_labels') if not args.ignored_labels else args.ignored_labels
        configs['ignored_labels'] = ignored_labels.split('|')
    else:
        configs['ignored_labels'] = []

    if config.has_option('settings', 'head') or args.head:
        root_whitelist = config.get('settings',
                                    'head') if not args.head else args.head
        configs['root_whitelist'] = root_whitelist.split('|')
    else:
        configs['root_whitelist'] = []

    if config.has_option('settings', 'query') or args.query:
        configs['query'] = (config.get('settings', 'query') if not args.query else args.query)
        # query has priority over the size, so we set it to 0
        configs['display_size'] = '0'
        configs['tree_size'] = '0'

    if args.compare:
        configs['compare'] = args.compare
    else:
        configs['compare'] = config.get('settings', 'compare') if config.has_option('settings', 'compare') else None

    configs['frequency_threshold'] = config.getfloat('settings', 'frequency_threshold', fallback=0) \
        if not args.frequency_threshold else args.frequency_threshold
    configs['lines_threshold'] = config.getint('settings', 'max_lines', fallback=0) \
        if not args.max_lines else args.max_lines

    configs['continuation_processing'] = config.getboolean('settings', 'continuation_processing', fallback=False) \
        if not args.continuation_processing else args.continuation_processing == 'yes'

    configs['grew_match'] = config.getboolean('settings',
                                              'grew_match') if not args.grew_match else args.grew_match == 'yes'
    configs['example'] = config.getboolean('settings', 'example') \
        if not args.example else args.example == 'yes'
    configs['label_subtypes'] = config.getboolean('settings', 'label_subtypes') \
        if not args.label_subtypes else args.label_subtypes == 'yes'

    if args.sentence_count_file:
        configs['sentence_count_file'] = args.sentence_count_file
    else:
        configs['sentence_count_file'] = config.get('settings', 'sentence_count_file') \
            if config.has_option('settings', 'sentence_count_file') else None

    if args.detailed_results_file:
        configs['detailed_results_file'] = args.detailed_results_file
    else:
        configs['detailed_results_file'] = config.get('settings', 'detailed_results_file') \
            if config.has_option('settings', 'detailed_results_file') else None

    if args.annodoc_example_dir:
        configs['annodoc_example_dir'] = args.annodoc_example_dir
    else:
        configs['annodoc_example_dir'] = config.get('settings', 'annodoc_example_dir') \
            if config.has_option('settings', 'annodoc_example_dir') else None

    if args.annodoc_detailed_dir:
        configs['annodoc_detailed_dir'] = args.annodoc_detailed_dir
    else:
        configs['annodoc_detailed_dir'] = config.get('settings', 'annodoc_detailed_dir') \
            if config.has_option('settings', 'annodoc_detailed_dir') else None

    configs['depsearch'] = config.getboolean('settings', 'depsearch') \
        if not args.depsearch else args.depsearch == 'yes'

    configs['nodes_number'] = True

    if configs['compare'] is not None:
        configs['other_input_path'] = configs['compare']

    # when processing_size is not given and display_size is set processing_size or throw an error!
    if configs['tree_size'] == '0' and configs['display_size'] != '0':
        if configs['complete_tree_type'] and configs['greedy_counter']:
            configs['tree_size'] = '1-10000000'
        else:
            raise ValueError('You have to specify `processing_size` parameter!')

    if configs['tree_size'] != '0' and configs['display_size'] == '0':
        raise ValueError('When `processing_size` setting is set, `size` has to be specified as well!')

    return configs


def read_settings(config_file, args=EmptyNamespace()):
    """
    Reads configuration file and merges it with arguments.
    :param config_file: path to config file.
    :param args: Namespace object
    :return:
    """
    config = configparser.ConfigParser()
    config.read(config_file)

    return read_configs(config, args)


def run(configs):
    """
    Executes STARK processing.
    :param configs: Dictionary containing execution settings.
    :return: Either object with results or None, when results are stored into tsv file.
    """

    filters = read_filters(configs)
    summary = count_subtrees(configs, filters)

    other_summary = None
    if configs['compare'] is not None:
        configs['input_path'] = configs['other_input_path']
        other_summary = count_subtrees(configs, filters)
    if configs['output']:
        writer = TSVWriter(summary, other_summary, filters, configs)
    else:
        writer = ObjectWriter(summary, other_summary, filters, configs)
    return writer.write()



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gc
import logging
import os

import pyconll

from stark.data.document import Document
from stark.data.processing.greedy_tree import GreedyTree
from stark.data.processing.query_tree import QueryTree
from stark.processing.cache import DocumentCache

logger = logging.getLogger('stark')


class DocumentProcessor(object):
    """
    A class that processes document.
    """
    def __init__(self, path, processor):
        self.path = path
        self.processor = processor
        self.cache = DocumentCache(self, path)

    def form_trees(self, summary, configs):
        """
        Forms trees by trying to load it from cache or processing it.
        :param summary:
        :return:
        """
        return self.cache.create_trees(summary, configs)

    @staticmethod
    def _reform_misc(misc):
        """
        Uses pyconll to reform misc column from dict created by the same library

        :param misc:
        :return:
        """
        return pyconll.unit.token._dict_mixed_conll_map(misc, '_', '|',
                                                 '=', ',',
                                                 lambda pair: pair[0].lower())

    def create_trees(self, summary, configs):
        """
        Creates trees based on configs and stores them in Document object.
        :param configs:
        :param summary:
        :return:
        """
        document = Document()

        logger.info("Reading file: " + self.path)
        train = pyconll.load_from_file(self.path)
        for sentence in train:
            token_nodes = []
            tokens = []
            for token in sentence:
                if not token.id.isdigit():
                    continue

                token_form = token.form if token.form is not None else '_'
                token_lemma = token.lemma if token.lemma is not None else '_'
                token_upos = token.upos if token.upos is not None else '_'
                token_xpos = token.xpos if token.xpos is not None else '_'
                token_misc = self._reform_misc(token.misc) if token.misc else '_'
                token_deprel = token.deprel if self.processor.configs['label_subtypes'] \
                    else token.deprel.split(':')[0]
                if self.processor.configs['greedy_counter']:
                    node = GreedyTree(int(token.id), token_form, token_lemma, token_upos, token_xpos, token_deprel,
                                      token.head, token.feats, token_misc, document, summary)
                else:
                    node = QueryTree(int(token.id), token_form, token_lemma, token_upos, token_xpos, token_deprel,
                                     token.head, token.feats, token_misc, document, summary)
                token_nodes.append(node)
                space_after = token.misc[
                                  'SpaceAfter'].pop() != 'No' if token.misc is not None and 'SpaceAfter' in token.misc \
                    else True
                tokens.append((token_form, space_after))

                summary.corpus_size += 1
            document.sentence_statistics.append({'id': sentence.id, 'tokens': tokens, 'count': {}})
            roots = []
            for token_id, token in enumerate(token_nodes):
                if isinstance(token.parent, int) or token.parent == '':
                    logger.warning('No parent: ' + sentence.id)
                    break
                if int(token.parent) == 0:
                    token.set_parent(None)
                    # add a conllu string if necessary
                    if configs['annodoc_example_dir'] is not None:
                        token.add_conll_sentence(sentence.conll())
                    roots.append(token)
                else:
                    parent_id = int(token.parent) - 1
                    if token_nodes[parent_id].children_split == -1 and token_id > parent_id:
                        token_nodes[parent_id].children_split = len(token_nodes[parent_id].children)
                    token_nodes[parent_id].add_child(token)
                    token.set_parent(token_nodes[parent_id])

            for token in token_nodes:
                if token.children_split == -1:
                    token.children_split = len(token.children)

            if not roots:
                logger.warning('No root: ' + sentence.id)

            document.trees.append(roots)

        del train
        gc.collect()

        return document



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stark.utils import create_output_string_deprel, create_output_string_lemma, create_output_string_upos, \
    create_output_string_xpos, create_output_string_feats, create_output_string_form, create_output_string_none, \
    create_output_string_misc

ROOT_WHITELIST_OPTIONS = ['deprel', 'feats', 'form', 'lemma', 'upos', 'misc']


def read_filters(configs):
    """
    Prepares filters for faster execution.
    :param configs:
    :return:
    """
    tree_size = configs['tree_size']
    tree_size_range = tree_size.split('-')
    tree_size_range = [int(r) for r in tree_size_range]

    display_size = configs['display_size']
    display_size_range = display_size.split('-')
    display_size_range = [int(r) for r in display_size_range]

    # set filters
    node_type = configs['node_type']
    if node_type:
        node_types = node_type.split('+')
        create_output_string_functs = []
        for node_type in node_types:
            assert node_type in ['deprel', 'lemma', 'upos', 'xpos', 'form', 'feats', 'misc'], \
                '"node_type" is not set up correctly'
            if node_type == 'deprel':
                create_output_string_funct = create_output_string_deprel
            elif node_type == 'lemma':
                create_output_string_funct = create_output_string_lemma
            elif node_type == 'upos':
                create_output_string_funct = create_output_string_upos
            elif node_type == 'xpos':
                create_output_string_funct = create_output_string_xpos
            elif node_type == 'feats':
                create_output_string_funct = create_output_string_feats
            elif node_type == 'misc':
                create_output_string_funct = create_output_string_misc
            else:
                create_output_string_funct = create_output_string_form
            create_output_string_functs.append(create_output_string_funct)
    else:
        create_output_string_functs = [create_output_string_none]
        node_types = ['generic']

    filters = {
        'create_output_string_functs': create_output_string_functs,
        'node_types': node_types,
        'tree_size_range': tree_size_range,
        'display_size_range': display_size_range,
        'cpu_cores': configs['cpu_cores'],
        'internal_saves': configs['internal_saves'],
        'input': configs['input_path'],
        'node_order': configs['node_order'],
        'dependency_type': configs['dependency_type'],
        'label_whitelist': configs['label_whitelist'],
        'ignored_labels': configs['ignored_labels'],
        'example': configs['example'],
        'sentence_count_file': configs['sentence_count_file'],
        'detailed_results_file': configs['detailed_results_file'],
        'annodoc': configs['annodoc_example_dir'],
        'complete_tree_type': configs['complete_tree_type'],
        'association_measures': configs['association_measures'],
        'nodes_number': configs['nodes_number'],
        'frequency_threshold': configs['frequency_threshold'],
        'lines_threshold': configs['lines_threshold'],
        'head_info': configs['head_info']
    }

    if configs['root_whitelist']:
        filters['root_whitelist'] = []

        for option in configs['root_whitelist']:
            attribute_dict = {}
            for attribute in option.split('&'):
                value = attribute.split('=')
                value[1] = '='.join(attribute.split('=')[1:])
                # look for negation sign, if there is only one character - ! it will search for form ! instead
                if value[0][0] == '!' and len(value[0]) > 1:
                    negation = True
                    value[0] = value[0][1:]
                else:
                    negation = False
                if len(value) == 1:
                    attribute_dict['form'] = (negation, value[0])
                else:
                    attribute_dict[value[0]] = (negation, value[1])
            filters['root_whitelist'].append(attribute_dict)
    else:
        filters['root_whitelist'] = []

    return filters


class Filter(object):
    @staticmethod
    def check_representation_tree(tree, filters):
        """
        Checks if greedy representation tree passes filters.
        :param filters:
        :param tree:
        :return:
        """
        return (
                Filter.check_tree_size(tree.tree_size, filters)
                and Filter.check_root_whitelist(tree.node.node.form, tree.node.node.lemma, tree.node.node.upos,
                                                tree.node.node.feats, tree.node.node.deprel, tree.node.node.misc, filters)
        )

    @staticmethod
    def check_query_tree(query_tree, form, lemma, upos, xpos, feats, deprel, children, filters):
        """
        Checks if attributes of a tree fit attributes of query_tree.
        :param query_tree:
        :param form:
        :param lemma:
        :param upos:
        :param xpos:
        :param feats:
        :param deprel:
        :param children:
        :param filters:
        :return:
        """
        filter_passed = (not filters['complete_tree_type'] or (len(children) == 0 and 'children' not in query_tree) or
                         ('children' in query_tree and len(children) == len(query_tree['children'])))

        # restrictions might not be in query tree when dealing with query counter and size
        if 'restrictions' not in query_tree or not query_tree['restrictions']:
            return filter_passed

        # ('deprel' not in option or (option['deprel'][1] == deprel) == (not option['deprel'][0])) and \
        for option in query_tree['restrictions']:
            filter_passed = ('form' not in option or (option['form'][1] == form) == (not option['form'][0])) and \
                ('lemma' not in option or (option['lemma'][1] == lemma) == (not option['lemma'][0])) and \
                ('upos' not in option or (option['upos'][1] == upos) == (not option['upos'][0])) and \
                ('xpos' not in option or (option['xpos'][1] == xpos) == (not option['xpos'][0])) and \
                ('deprel' not in option or (option['deprel'][1] == deprel) == (not option['deprel'][0])) and \
                Filter._check_query_tree_feats(option, feats)

            if filter_passed:
                return True

        return False

    @staticmethod
    def _check_query_tree_feats(option, feats):
        """
        Checks if feats of a tree fit feats of option.
        :param option:
        :param feats:
        :return:
        """
        if 'feats_detailed' not in option:
            return True

        for feat in option['feats_detailed'].keys():
            if feat not in feats:
                if not option['feats_detailed'][feat][0]:
                    return False
            elif option['feats_detailed'][feat][1] != feats[feat] and not option['feats_detailed'][feat][0]:
                return False
            elif option['feats_detailed'][feat][1] == feats[feat] and option['feats_detailed'][feat][0]:
                return False
        return True

    @staticmethod
    def check_tree_size(size, filters):
        """
        Checks if tree size is in filtered range.
        :param size:
        :param filters:
        :return:
        """
        return filters['tree_size_range'][0] <= size <= filters['tree_size_range'][-1]

    @staticmethod
    def check_root_whitelist(form, lemma, upos, feats, deprel, misc, filters):
        """
        When root whitelist exists checks if element parameters are acceptable.
        :param form:
        :param lemma:
        :param upos:
        :param feats:
        :param deprel:
        :param filters:
        :return:
        """
        if not filters['root_whitelist']:
            return True

        for option in filters['root_whitelist']:
            filter_passed = True
            # check if attributes are valid
            for key in option.keys():
                if key not in ROOT_WHITELIST_OPTIONS:
                    # key not in feats and not negated
                    if key not in feats:
                        if not option[key][0]:
                            filter_passed = False
                    # key different, then in head and not negated
                    elif option[key][1] != feats[key] and not option[key][0]:
                        filter_passed = False
                    # key equal, and negated
                    elif option[key][1] == feats[key] and option[key][0]:
                        filter_passed = False

            filter_passed = filter_passed and \
                ('deprel' not in option or (option['deprel'][1] == deprel) == (not option['deprel'][0])) and \
                ('form' not in option or (option['form'][1] == form) == (not option['form'][0])) and \
                ('lemma' not in option or (option['lemma'][1] == lemma) == (not option['lemma'][0])) and \
                ('upos' not in option or (option['upos'][1] == upos) == (not option['upos'][0])) and \
                ('misc' not in option or (option['misc'][1] == misc) == (not option['misc'][0]))

            if filter_passed:
                return True

        return False

    @staticmethod
    def check_label_whitelist(deprel, filters):
        """
        When label whitelist exists, check if deprel is in it.
        :param deprel:
        :param filters:
        :return:
        """
        return not filters['label_whitelist'] or deprel in filters['label_whitelist']



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stark.data.representation.tree import RepresentationTree


class QueryRepresentationTree(RepresentationTree):
    def __init__(self, node, children, filters):
        super().__init__(node, children)

    @classmethod
    def copy(cls, node, children, filters):
        """
        Copies original object with empty children.
        :param filters:
        :param children:
        :param node:
        :return:
        """
        return cls(node, children, filters)



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy

from stark.data.representation.node import RepresentationNode
from stark.data.representation.query_tree import QueryRepresentationTree
from stark.data.processing.tree import Tree
from stark.processing.filters import Filter


class QueryTree(Tree):
    def __init__(self, index, form, lemma, upos, xpos, deprel, head, feats_detailed, token_misc, document, summary):
        super().__init__(index, form, lemma, upos, xpos, deprel, head, feats_detailed, token_misc, document, summary)

    def get_subtrees(self, permanent_query_trees, temporary_query_trees, filters):
        """

        :param filters:
        :param permanent_query_trees:
        :param temporary_query_trees:
        """

        # list of all children queries grouped by parent queries
        all_query_indices = []

        active_permanent_query_trees = []
        for permanent_query_tree in permanent_query_trees:
            if (self._fits_static_requirements(permanent_query_tree, filters)
                    and self._fits_permanent_requirements(filters)):
                active_permanent_query_trees.append(permanent_query_tree)
                if 'children' in permanent_query_tree:
                    all_query_indices.append((permanent_query_tree['children'], True))
                    # r_all_query_indices.append((permanent_query_tree['r_children'], True))

        active_temporary_query_trees = []
        successful_temporary_queries = []
        for i, temporary_query_tree in enumerate(temporary_query_trees):
            if (self._fits_static_requirements(temporary_query_tree, filters)
                    and self._fits_temporary_requirements(filters)):
                active_temporary_query_trees.append(temporary_query_tree)
                successful_temporary_queries.append(i)
                if 'children' in temporary_query_tree:
                    all_query_indices.append((temporary_query_tree['children'], False))

        partial_subtrees, complete_answers = QueryTree._get_all_query_indices(len(temporary_query_trees),
                                                                         len(permanent_query_trees),
                                                                         permanent_query_trees,
                                                                         all_query_indices, self.children,
                                                                         filters)

        merged_partial_answers = []
        i_question = 0
        # i_child is necessary, because some queries may be answered at the beginning and were not passed to children.
        # i_child is used to point where we are inside answers
        i_answer = 0
        # go over all permanent and temporary query trees
        while i_question < len(active_permanent_query_trees) + len(active_temporary_query_trees):
            # permanent query trees always have left and right child
            i_answer = self._order_dependent_queries(active_permanent_query_trees, active_temporary_query_trees,
                                                    partial_subtrees, merged_partial_answers,
                                                    i_question, i_answer, filters)

            i_question += 1

        for i in range(len(active_permanent_query_trees)):
            # erase first and last brackets when adding new query result
            add_subtree = [subtree.ignore_labels(filters) for subtree in merged_partial_answers[i]]
            complete_answers[i].extend(add_subtree)

        # answers to valid queries
        partial_answers = [[] for _ in range(len(temporary_query_trees))]
        for inside_i, outside_i in enumerate(successful_temporary_queries):
            partial_answers[outside_i] = merged_partial_answers[
                len(active_permanent_query_trees) + inside_i]

        return partial_answers, complete_answers

    def _fits_static_requirements_feats(self, query_tree):
        if 'feats_detailed' not in query_tree:
            return True

        for feat in query_tree['feats_detailed'].keys():
            if (feat not in self.feats_detailed or
                    query_tree['feats_detailed'][feat] != next(iter(self.feats_detailed[feat].values()))):
                return False

        return True

    def _fits_permanent_requirements(self, filters):
        return Filter.check_root_whitelist(self.form, self.lemma, self.upos,
                                           self.feats, self.deprel, self.misc, filters)

    def _fits_temporary_requirements(self, filters):
        return Filter.check_label_whitelist(self.deprel, filters)

    def _fits_static_requirements(self, query_tree, filters):
        return Filter.check_query_tree(query_tree, self.form, self.lemma, self.upos,
                                       self.xpos, self.feats, self.deprel,
                                       self.children, filters)


    @staticmethod
    def _merge_results3(child, new_results, filters):
        if filters['node_order']:
            new_child = child

        else:
            new_child = sorted(child, key=lambda x: x[0].get_key(filters))

        children_groups = []

        for i_answer, answer in enumerate(new_child):
            children_groups = QueryTree._create_children_groups(children_groups, [[answer_part] for answer_part in answer])

        results = []
        for result in new_results:
            for children in children_groups:
                new_result = copy(result)
                new_result.set_children(children)
                results.append(new_result)

        return results

    @staticmethod
    def _get_all_query_indices(temporary_query_nb, permanent_query_nb, permanent_query_trees, all_query_indices,
                               children, filters):
        partial_answers = [[] for _ in range(permanent_query_nb + temporary_query_nb)]
        complete_answers = [[] for _ in range(permanent_query_nb)]

        # list of pairs (index of query in group, group of query)
        child_queries = [all_query_indice[0] for all_query_indice in all_query_indices]

        answers_lengths = [len(query) for query in child_queries]

        child_queries_flatten = [query_part for query in child_queries for query_part in query]

        all_new_partial_answers = [[] for _ in child_queries_flatten]

        child_queries_flatten_dedup = []
        child_queries_flatten_dedup_indices = []
        for query_part in child_queries_flatten:
            try:
                index = child_queries_flatten_dedup.index(query_part)
            except ValueError:
                index = len(child_queries_flatten_dedup)
                child_queries_flatten_dedup.append(query_part)

            child_queries_flatten_dedup_indices.append(index)

        # ask children all queries/partial queries
        for child in children:
            # obtain children results
            new_partial_answers_dedup, new_complete_answers = child.get_subtrees(permanent_query_trees,
                                                                                 child_queries_flatten_dedup,
                                                                                 filters)

            assert len(new_partial_answers_dedup) == len(child_queries_flatten_dedup)

            # duplicate results again on correct places
            for i, flattened_index in enumerate(child_queries_flatten_dedup_indices):
                all_new_partial_answers[i].append(new_partial_answers_dedup[flattened_index])

            for i in range(len(new_complete_answers)):
                # TODO add order rearagement (TO KEY)
                complete_answers[i].extend(new_complete_answers[i])

        # merge answers in appropriate way
        i = 0
        # iterate over all answers per queries
        for answer_i, answer_length in enumerate(answers_lengths):
            # iterate over answers of query
            partial_answers[answer_i] = QueryTree._create_answers(all_new_partial_answers[i:i + answer_length],
                                                             answer_length, filters)
            i += answer_length

        return partial_answers, complete_answers

    def _order_dependent_queries(self, active_permanent_query_trees, active_temporary_query_trees, partial_subtrees,
                                 merged_partial_subtrees, i_query, i_answer, filters):
        node = RepresentationNode(self, self.index, filters['create_output_string_functs'])

        if i_query < len(active_permanent_query_trees):
            if 'children' in active_permanent_query_trees[i_query]:
                merged_partial_subtrees.append(
                    self._create_output_children(partial_subtrees[i_answer],
                                                 [QueryRepresentationTree(node, [], filters)], filters))
                i_answer += 1
            else:
                merged_partial_subtrees.append([QueryRepresentationTree(node, [], filters)])
        else:
            if 'children' in active_temporary_query_trees[i_query - len(active_permanent_query_trees)]:
                merged_partial_subtrees.append(
                    self._create_output_children(partial_subtrees[i_answer],
                                                 [QueryRepresentationTree(node, [], filters)], filters))
                i_answer += 1
            else:
                merged_partial_subtrees.append([QueryRepresentationTree(node, [], filters)])
        return i_answer

    @staticmethod
    def _create_children_groups(left_parts, right_parts):
        if not left_parts:
            return right_parts

        if not right_parts:
            return left_parts

        all_children_group_possibilities = []
        for left_part in left_parts:
            for right_part in right_parts:
                new_part = copy(left_part)
                new_part.extend(right_part)
                all_children_group_possibilities.append(new_part)
        return all_children_group_possibilities

    def _create_output_children(self, children, new_results, filters):
        merged_results = []
        for i_child, child in enumerate(children):
            merged_results.extend(self._merge_results3(child, new_results, filters))
        return merged_results

    @staticmethod
    def _create_answers(separated_answers, answer_length, filters):
        partly_built_trees = [[None] * answer_length]
        partly_built_trees_architecture_indices = [[None] * answer_length]
        built_trees = []
        built_trees_architecture_indices = []

        # iterate over children first, so that new partly built trees are added only after all results of specific
        # child are added
        for child_i in range(len(separated_answers[0])):
            new_partly_built_trees = []
            new_partly_built_trees_architecture_indices = []
            # iterate over answers parts
            for answer_part_i in range(len(separated_answers)):
                # necessary because some parts do not pass filters and are not added
                if separated_answers[answer_part_i][child_i]:
                    for tree_part_i, tree_part in enumerate(partly_built_trees):
                        if not tree_part[answer_part_i]:
                            new_tree_part = copy(tree_part)
                            new_tree_part_architecture_indices = (
                                copy(partly_built_trees_architecture_indices[tree_part_i]))
                            new_tree_part[answer_part_i] = separated_answers[answer_part_i][child_i]
                            new_tree_part_architecture_indices[answer_part_i] = child_i
                            completed_tree_part = True
                            for val_i, val in enumerate(new_tree_part):
                                if not val:
                                    completed_tree_part = False
                            if completed_tree_part:
                                built_trees.append(new_tree_part)
                                built_trees_architecture_indices.append(new_tree_part_architecture_indices)
                            else:
                                new_partly_built_trees.append(new_tree_part)
                                new_partly_built_trees_architecture_indices.append(new_tree_part_architecture_indices)
                        else:
                            # pass over repetitions of same words
                            pass

            partly_built_trees.extend(new_partly_built_trees)
            partly_built_trees_architecture_indices.extend(new_partly_built_trees_architecture_indices)

        l_ordered_built_trees, unique_trees_architecture = [], []

        if built_trees:
            # sort 3 arrays by architecture indices
            temp_trees_index, temp_trees = (list(t) for t in zip(
                *sorted(zip(built_trees_architecture_indices, built_trees))))

            # order outputs and erase duplicates
            for tree, tree_index in zip(temp_trees, temp_trees_index):
                new_tree_index, new_tree = (list(t) for t in zip(*sorted(zip(tree_index, tree))))
                is_unique = True
                for unique_tree in unique_trees_architecture:
                    already_in = True
                    for part_i in range(len(unique_tree)):
                        if (len(unique_tree[part_i]) != len(new_tree[part_i])
                                or any(unique_tree[part_i][i_unique_part].get_order_key(filters) !=
                                       new_tree[part_i][i_unique_part].get_order_key(filters)
                                       for i_unique_part in range(len(unique_tree[part_i])))):
                            already_in = False
                            break
                    if already_in:
                        is_unique = False
                        break

                if is_unique:
                    unique_trees_architecture.append(new_tree)
                    l_ordered_built_trees.append(new_tree)
        return l_ordered_built_trees



UNIVERSAL_FEATURES = {
    "PronType", "NumType", "Poss", "Reflex", "Foreign", "Abbr", "Gender",
    "Animacy", "Number", "Case", "Definite", "Degree", "VerbForm", "Mood",
    "Tense", "Aspect", "Voice", "Evident", "Polarity", "Person", "Polite"
}


import re
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# read the version from classla/_version.py
version_file_contents = open(path.join(here, 'stark/_version.py'), encoding='utf-8').read()
VERSION = re.compile('__stark_version__ = \'(.*)\'').search(version_file_contents).group(1)

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

setup(name='stark-trees',
  version=VERSION,
  description=u"Parser for dependency trees",
  long_description=long_description,
  long_description_content_type="text/markdown",
  author='CLARIN.SI',
  author_email='info@clarin.si',
  license='Apache 2.0',
  packages=find_packages(),
  include_package_data=True,
  install_requires=[
    'pyconll>=3.1.0',
    'networkx>=3.3',
    'tqdm>=4.66.4'
  ],
  entry_points={
      "console_scripts": [
      "stark=stark.cli:main",  # This maps `stark` to `stark.cli.main()`
    ],
  },
  # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
  classifiers=[
      # How mature is this project? Common values are
      #   3 - Alpha
      #   4 - Beta
      #   5 - Production/Stable
      'Development Status :: 4 - Beta',

      # Indicate who your project is intended for
      'Intended Audience :: Developers',
      'Intended Audience :: Education',
      'Intended Audience :: Science/Research',
      'Intended Audience :: Information Technology',
      'Topic :: Scientific/Engineering',
      'Topic :: Scientific/Engineering :: Information Analysis',
      'Topic :: Text Processing',
      'Topic :: Text Processing :: Filters',
      'Topic :: Text Processing :: Linguistic',
      'Topic :: Software Development',
      'Topic :: Software Development :: Libraries',

      # Specify the Python versions you support here. In particular, ensure
      # that you indicate whether you support Python 2, Python 3 or both.
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',
      'Programming Language :: Python :: 3.8',
      'Programming Language :: Python :: 3.9',
      'Programming Language :: Python :: 3.10',
  ],
  python_requires='>=3.6',
  # What does your project relate to?
  keywords='natural-language-processing nlp natural-language-understanding dependency-trees extraction comparison clarinsi',

)



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stark.data.processing.tree import Tree
from stark.data.representation.greedy_tree import GreedyRepresentationTree
from stark.data.representation.node import RepresentationNode
from stark.processing.filters import Filter


class GreedyTree(Tree):
    def __init__(self, index, form, lemma, upos, xpos, deprel, head, feats_detailed, token_misc, document, summary):
        super().__init__(index, form, lemma, upos, xpos, deprel, head, feats_detailed, token_misc, document, summary)

    @staticmethod
    def _processing_filter(combinations, child_active_tree, filters):
        """
        Utilizes loose filters that might improve greedy performance (ie. tree_size).
        :param combinations: List containing previous children combinations.
        :param child_active_tree: A child to be added to combinations if it passes filters.
        :param filters:
        :return:
        """
        # checks whether tree size is not too high
        # checks if label is in white list (if such a list exists)
        return (combinations[0] + child_active_tree.tree_size <= filters['tree_size_range'][-1] and
                Filter.check_label_whitelist(child_active_tree.node.node.deprel, filters))

    @staticmethod
    def _merge_incomplete_combinations(combinations, child_active_trees, filters):
        """
        Creates all possible combinations of children trees when complete_tree_type=no. It utilizes loose processing
        filters.
        :param combinations: A list of all combinations of a tree node.
        :param child_active_trees: A list of trees that contain child node.
        :param filters:
        :return:
        """
        # create all viable children combinations
        new_active_trees = []

        for child_active_tree in child_active_trees:
            for combination in combinations:
                if GreedyTree._processing_filter(combination, child_active_tree, filters):
                    new_active_trees.append((combination[0] + child_active_tree.tree_size,
                                             combination[1] + [child_active_tree]))

        combinations.extend(new_active_trees)
        return combinations

    @staticmethod
    def _merge_complete_combinations(combinations, child_active_trees, filters):
        """
        Creates all possible combinations of children trees when complete_tree_type=yes. Always generates all complete
        trees (there are fairly few of them) - no filtering!
        :param combinations:
        :param child_active_trees:
        :return:
        """
        if not child_active_trees or not combinations or not GreedyTree._processing_filter(combinations[0], child_active_trees[0], filters):
            return []
        new_combinations_size = combinations[0][0] + child_active_trees[0].tree_size
        combinations[0][1].append(child_active_trees[0])
        combinations[0] = (new_combinations_size, combinations[0][1])
        return combinations

    @staticmethod
    def _merge_combinations(combinations, child_active_trees, filters):
        """
        Creates all possible combinations of children trees.
        :param combinations: A list of all combinations of a tree node.
        :param child_active_trees: A list of trees that contain child node.
        :param filters:
        :return:
        """
        if filters['complete_tree_type']:
            combinations = GreedyTree._merge_complete_combinations(combinations, child_active_trees, filters)
        else:
            combinations = GreedyTree._merge_incomplete_combinations(combinations, child_active_trees, filters)

        return combinations

    def get_subtrees(self, filters):
        """
        A recursion that builds representation trees (representation_tree) and stores and collect them (trees).
        :param filters:
        :return:
        """
        # a list that stores all trees
        trees = []
        # A tuple containing cumulative size of all children, coupled with a list of children combinations.
        # The list contains combinations that are connected to the current node (and are relevant for further
        # generation). Set to [[]] because you always want an empty tree containing only itself.
        combinations = [(1, [])]

        node = RepresentationNode(self, self.index, filters['create_output_string_functs'])
        for child in self.children:
            child_active_trees, child_trees = child.get_subtrees(filters)
            combinations = GreedyTree._merge_combinations(combinations, child_active_trees, filters)
            trees.extend(child_trees)

        active_trees = []
        for combination in combinations:
            active_trees.append(GreedyRepresentationTree(node, combination, filters))
        trees.extend(active_trees)
        return active_trees, trees



# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os

import logging
from pathlib import Path

from stark.data.document import Document
from stark.data.summary import Summary
from stark.utils import load_zipped_pickle, save_zipped_pickle

logger = logging.getLogger('stark')


class ProcessorCache(object):
    """
    Caching class, used as a wrapper for processing multiple files. It enables continuation processing.
    """
    def __init__(self, processor):
        self.configs = processor.configs
        self.already_processed = set()
        self._checkpoint_path = Path(self.configs['internal_saves'], 'checkpoint.pkl') \
            if self.configs['internal_saves'] is not None else None
        self.processor = processor

    def load_cache(self, summary):
        """
        Loads cache when applicable and deletes it if needed.
        :return:
        summary: Either received summary or loaded one.
        """
        if (self._checkpoint_path is None or not self._checkpoint_path.exists() or
                not self.configs['continuation_processing']):
            if self._checkpoint_path is not None and self._checkpoint_path.exists():
                os.remove(self._checkpoint_path)
        else:
            return self._load_cache()
        return summary

    def _load_cache(self):
        """
        Actually loads cache.
        :return:
        summary: Loaded summary
        """
        self.already_processed, summary_data = load_zipped_pickle(self._checkpoint_path)
        summary = Summary.create_summary_from_cache(summary_data)
        return summary

    def _save_cache(self, summary):
        """
        Actually saves cache.
        :return:
        """
        summary_data = summary.get_summary_data()
        save_zipped_pickle(
            (self.already_processed, summary_data),
            self._checkpoint_path, protocol=2)

    def process_trees(self, path, summary):
        """
        Skips documents that were already processed and stores processed results in cache.
        :param summary: Processing summary.
        :param path: Path to dir.
        :return:
        summary: A collection of datapoints used for result generation.
        """
        if str(path) in self.already_processed:
            logger.info(f'Skipping: {str(path)}')
            return summary

        summary = self.processor.run(path, summary)
        self.already_processed.add(str(path))
        if self._checkpoint_path:
            self._save_cache(summary)

        return summary


class DocumentCache(object):
    """
    Cache that stores generated sentence trees from files. When used, reading input files is not necessary anymore.
    May be problematic if users modify input files and don't rename them.
    """
    def __init__(self, document_processor, path):
        configs = document_processor.processor.configs
        self.document_processor = document_processor
        self.path = path
        self._internal_file = os.path.join(configs['internal_saves'], hashlib.sha1(path.encode('utf-8')).hexdigest()) \
            if configs['internal_saves'] is not None else None
        # do not save cache if input is dir
        self._save = not os.path.isdir(configs['input_path'])

    def create_trees(self, summary, configs):
        if self._internal_file is None or not os.path.exists(self._internal_file) or not self._save:
            document = self.document_processor.create_trees(summary, configs)

            if self._internal_file is not None and self._save:
                self._save_cache(document, summary)
        else:
            document = self._load_cache(summary)

        return document

    def _save_cache(self, document, summary):
        document_data = document.get_document_data() + [summary.corpus_size, summary.feats_dict]
        save_zipped_pickle(document_data, self._internal_file, protocol=2)

    def _load_cache(self, summary):
        document_data = load_zipped_pickle(self._internal_file)
        summary.corpus_size, summary.feats_dict = document_data[-2:]
        return Document.create_document_from_cache(document_data[:-2])



import sys
import time

import stark
import logging
from stark.stark import read_settings, parse_args
logger = logging.getLogger('stark')


def main():
    args = parse_args(sys.argv[1:])

    settings = read_settings(args.config_file, args)

    stark.run(settings)


if __name__ == "__main__":
    start_time = time.time()
    main()
    logger.info("Total:")
    logger.info("--- %s seconds ---" % (time.time() - start_time))



# Copyright 2019 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import string


class RepresentationTree(object):
    def __init__(self, node, children):
        self.node = node
        self.children = children

    @classmethod
    @abc.abstractmethod
    def copy(cls, node, children, filters):
        return

    def set_children(self, children):
        self.children = children

    def get_grew(self):
        nodes = [self.node]
        links = []

        if self.children:
            for child in self.children:
                links.append((self.node, child.node))
                c_nodes, c_links = child.get_grew()
                nodes.extend(c_nodes)
                links.extend(c_links)

        return nodes, links

    def ignore_nodes(self, filters):
        """
        Drops nodes in result tree, that are supposed to be ignored.
        """
        children = [child for child in self.children if child.node.node.deprel not in filters['ignored_labels']]

        new_children = []
        for child in children:
            new_children.append(child.ignore_nodes(filters))

        return self.copy(self.node, new_children, filters)

    def get_key_array(self, filters):
        """
        A code that generates key and array of a tree simultaneously (for faster execution).
        :return:
        key: Key of a tree
        array: Array of tree elements
        """

        array = []
        key = ''
        write_self_node_to_result = False
        if self.children:
            children = self.children
            for child in children:
                if filters['node_order'] and child.node.location < self.node.location:
                    k, a = child.get_key_array(filters)
                    array += a

                    if filters['dependency_type']:
                        separator = ' <' + child.node.node.deprel + ' '
                    else:
                        separator = ' < '
                    key += k + separator
                else:
                    if not write_self_node_to_result:
                        write_self_node_to_result = True
                        key += self.node.name
                        array += [self.node.name_parts]
                    if filters['dependency_type']:
                        separator = ' >' + child.node.node.deprel + ' '
                    else:
                        separator = ' > '
                    k, a = child.get_key_array(filters)
                    array += a
                    key += separator + k

            if not write_self_node_to_result:
                key += self.node.name
                array += [self.node.name_parts]
            array = array
            key = '(' + key + ')'
        else:
            array = [self.node.name_parts]
            key = self.node.name
        return key, array

    def get_key(self, filters):
        """
        A code that returns and (if necessary generates) key of a tree. (used for `Tree` column in output)
        :return:
        key: Key of a tree
        """
        key = ''
        write_self_node_to_result = False
        if self.children:
            children = self.children
            for child in children:
                if filters['node_order'] and child.node.location < self.node.location:
                    if filters['dependency_type']:
                        separator = ' <' + child.node.node.deprel + ' '
                    else:
                        separator = ' < '
                    key += child.get_key(filters) + separator
                else:
                    if not write_self_node_to_result:
                        write_self_node_to_result = True
                        key += self.node.name
                    if filters['dependency_type']:
                        separator = ' >' + child.node.node.deprel + ' '
                    else:
                        separator = ' > '
                    key += separator + child.get_key(filters)

            if not write_self_node_to_result:
                key += self.node.name
            key = '(' + key + ')'
        else:
            key = self.node.name
        return key

    def get_key_sorted(self, filters):
        key = ''
        write_self_node_to_result = False
        if self.children:
            children = sorted(self.children, key=lambda x: x.node.name)
            for child in children:
                if not write_self_node_to_result:
                    write_self_node_to_result = True
                    key += self.node.name
                if filters['dependency_type']:
                    separator = ' >' + child.node.node.deprel + ' '
                else:
                    separator = ' > '
                key += separator + child.get_key_sorted(filters)

            if not write_self_node_to_result:
                key += self.node.name
            key = '(' + key + ')'
        else:
            key = self.node.name
        return key

    def get_order_key(self, filters):
        order_key = ''
        write_self_node_to_result = False
        if self.children:
            for child in self.children:
                if filters['node_order'] and child.node.location < self.node.location:
                    if filters['dependency_type']:
                        separator = ' <' + child.node.node.deprel + ' '
                    else:
                        separator = ' < '
                    order_key += child.get_order_key(filters) + separator
                else:
                    if not write_self_node_to_result:
                        write_self_node_to_result = True
                        order_key += str(self.node.location)
                    if filters['dependency_type']:
                        separator = ' >' + child.node.node.deprel + ' '
                    else:
                        separator = ' > '
                    order_key += separator + child.get_order_key(filters)
            if not write_self_node_to_result:
                order_key += str(self.node.location)
            order_key = '(' + order_key + ')'
        else:
            order_key = str(self.node.location)
        return order_key

    def get_order(self, filters):
        order = []
        write_self_node_to_result = False
        if self.children:
            for child in self.children:
                if filters['node_order'] and child.node.location < self.node.location:
                    order += child.get_order(filters)
                else:
                    if not write_self_node_to_result:
                        write_self_node_to_result = True
                        order += [self.node.location]
                    order += child.get_order(filters)

            if not write_self_node_to_result:
                order += [self.node.location]
            order = order
        else:
            order = [self.node.location]
        return order

    def get_array(self, filters):
        array = []
        write_self_node_to_result = False
        if self.children:
            for child in self.children:
                if filters['node_order'] and child.node.location < self.node.location:
                    array += child.get_array(filters)
                else:
                    if not write_self_node_to_result:
                        write_self_node_to_result = True
                        array += [self.node.name_parts]
                    array += child.get_array(filters)

            if not write_self_node_to_result:
                array += [self.node.name_parts]
            array = array
        else:
            array = [self.node.name_parts]
        return array

    @staticmethod
    def get_order_letters(order):
        order_letters = [''] * len(order)
        for i in range(len(order)):
            ind = order.index(min(order))
            order[ind] = 10000
            order_letters[ind] = string.ascii_uppercase[i % 26]
        return ''.join(order_letters)

    def ignore_labels(self, filters):
        if filters['ignored_labels']:
            return self.ignore_nodes(filters)
        return self

    def get_location_mapper(self, filters):
        mapper = {}
        location_array = self.get_array_location(filters)
        for i in range(len(location_array)):
            mapper[location_array[i]] = string.ascii_uppercase[i] if i < 26 else (string.ascii_uppercase[i % 26] +
                                                                                  str(int(i / 26)))

        return mapper

    def get_array_location(self, filters):
        array = []
        write_self_node_to_result = False
        if self.children:
            for child in self.children:
                if filters['node_order'] and child.node.location < self.node.location:
                    array += child.get_array_location(filters)
                else:
                    if not write_self_node_to_result:
                        write_self_node_to_result = True
                        array += [self.node.location]
                    array += child.get_array_location(filters)

            if not write_self_node_to_result:
                array += [self.node.location]
        else:
            array = [self.node.location]
        return array
